@@ -2,6 +2,10 @@
 Automount utility.
 """
 
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
+
 __all__ = ['AutoMounter']
 
 
@@ -24,25 +28,9 @@ class AutoMounter(object):
         :param Mounter mounter: mounter object
         """
         self._mounter = mounter
-        mounter.udisks.connect_all(self)
-
-    def device_added(self, device):
-        """
-        Mount newly added devices.
-
-        :param Device device: newly added device
-        """
-        if self._mounter.is_handleable(device):
-            self._mounter.add(device)
-
-    def media_added(self, device):
-        """
-        Mount newly added media.
-
-        :param Device device: device with newly added media
-        """
-        if self._mounter.is_handleable(device):
-            self._mounter.add(device)
+        mounter.udisks.connect('device_changed', self.device_changed)
+        mounter.udisks.connect('device_added', mounter.auto_add)
+        mounter.udisks.connect('media_added', mounter.auto_add)
 
     def device_changed(self, old_state, new_state):
         """
@@ -53,5 +41,7 @@ class AutoMounter(object):
         """
         # udisks2 sometimes adds empty devices and later updates them which
         # makes is_external become true not at device_added time:
-        if self._mounter.is_handleable(new_state) and not self._mounter.is_handleable(old_state):
-            self._mounter.add(new_state)
+        if (self._mounter.is_addable(new_state)
+                and not self._mounter.is_addable(old_state)
+                and not self._mounter.is_removable(old_state)):
+            self._mounter.auto_add(new_state)
