@@ -5,17 +5,26 @@ The application classes in this module are installed as executables via
 setuptools entry points.
 """
 
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
+# import udiskie.depend first - for side effects!
+from .depend import has_Notify, has_Gtk
+
 import inspect
 import logging
-import warnings
+import traceback
 
 from docopt import docopt, DocoptExit
 
+from gi.repository import GLib
+
 import udiskie
 import udiskie.config
 import udiskie.mount
-from udiskie.common import extend
-from udiskie.locale import _
+from .async_ import AsyncList, Coroutine, Return, RunForever
+from .common import extend, str2unicode
+from .locale import _
 
 
 __all__ = [
@@ -25,40 +34,36 @@ __all__ = [
 ]
 
 
-warnings.filterwarnings("ignore", ".*could not open display.*", Warning)
-warnings.filterwarnings("ignore", ".*g_object_unref.*", Warning)
-
-
-def get_backend(clsname, version=None):
+@Coroutine.from_generator_function
+def get_backend(version=None):
     """
     Return UDisks service.
 
-    :param str clsname: requested service object
     :param int version: requested UDisks backend version
     :returns: UDisks service wrapper object
-    :raises dbus.DBusException: if unable to connect to UDisks dbus service.
+    :raises GLib.GError: if unable to connect to UDisks dbus service.
     :raises ValueError: if the version is invalid
 
     If ``version`` has a false truth value, try to connect to UDisks1 and
     fall back to UDisks2 if not available.
     """
     if not version:
-        from udiskie.dbus import DBusException
         try:
-            return get_backend(clsname, 1)
-        except DBusException:
+            daemon = yield get_backend(2)
+        except GLib.GError:
             log = logging.getLogger(__name__)
-            log.warning(_('Failed to connect UDisks1 dbus service..\n'
-                          'Falling back to UDisks2 [experimental].'))
-            return get_backend(clsname, 2)
+            log.warning(_('Failed to connect UDisks2 dbus service..\n'
+                          'Falling back to UDisks1.'))
+            daemon = yield get_backend(1)
     elif version == 1:
         import udiskie.udisks1
-        return getattr(udiskie.udisks1, clsname)()
+        daemon = yield udiskie.udisks1.Daemon.create()
     elif version == 2:
         import udiskie.udisks2
-        return getattr(udiskie.udisks2, clsname)()
+        daemon = yield udiskie.udisks2.Daemon.create()
     else:
         raise ValueError(_("UDisks version not supported: {0}!", version))
+    yield Return(daemon)
 
 
 class Choice(object):
@@ -99,7 +104,7 @@ class Value(object):
 
     def __call__(self, args):
         """Get the value of the command line argument."""
-        return args[self._name]
+        return str2unicode(args[self._name])
 
 
 class OptionalValue(object):
@@ -111,7 +116,7 @@ class OptionalValue(object):
 
     def __call__(self, args):
         """Get the value of the command line argument."""
-        return self._choice(args) and args[self._name]
+        return self._choice(args) and str2unicode(args[self._name])
 
 
 class _EntryPoint(object):
@@ -174,7 +179,7 @@ class _EntryPoint(object):
         # initialize instance variables
         self.config = config
         self.options = options
-        self._init(config, options)
+        self.exit_status = 0
 
     def program_options(self, args):
         """
@@ -217,24 +222,41 @@ class _EntryPoint(object):
         """Get the name of the CLI utility."""
         raise NotImplementedError()
 
-    def _init(self, config, options):
+    def _init(self):
         """
         Fully initialize Daemon object.
 
-        :param Config config: configuration object
-        :param options: program options
+        :returns: the main application task
+        :rtype: Async
         """
         raise NotImplementedError()
 
     def run(self):
         """
-        Run main program logic.
+        Run the main loop.
 
-        :param options: program options
         :returns: exit code
         :rtype: int
         """
-        raise NotImplementedError()
+        self.mainloop = GLib.MainLoop()
+        self._start_async_tasks()
+        try:
+            self.mainloop.run()
+            return self.exit_status
+        except KeyboardInterrupt:
+            return 1
+
+    @Coroutine.from_generator_function
+    def _start_async_tasks(self):
+        """Start asynchronous operations."""
+        try:
+            self.udisks = yield get_backend(self.options['udisks_version'])
+            yield self._init()
+        except Exception:
+            self.exit_status = 1
+            # Print the stack trace only up to the current level:
+            traceback.print_exc()
+        self.mainloop.quit()
 
 
 class Daemon(_EntryPoint):
@@ -271,6 +293,12 @@ class Daemon(_EntryPoint):
         -s, --smart-tray                        Auto hide tray icon
         -T, --no-tray                           Disable tray icon
 
+        --password-cache MINUTES                Set password cache timeout
+        --no-password-cache                     Disable password cache
+
+        -p COMMAND, --password-prompt COMMAND   Command for password retrieval
+        -P, --no-password-prompt                Disable unlocking
+
     Deprecated options:
         -f PROGRAM, --file-manager PROGRAM      Set program for browsing
         -F, --no-file-manager                   Disable browsing
@@ -282,7 +310,9 @@ class Daemon(_EntryPoint):
         'automount': True,
         'notify': True,
         'tray': False,
-        'file_manager': 'xdg-open'
+        'file_manager': 'xdg-open',
+        'password_prompt': 'builtin:gui',
+        'password_cache': False,
     })
 
     option_rules = extend(_EntryPoint.option_rules, {
@@ -293,33 +323,72 @@ class Daemon(_EntryPoint):
             '--no-tray': False,
             '--smart-tray': 'auto'}),
         'file_manager': OptionalValue('--file-manager'),
+        'password_prompt': OptionalValue('--password-prompt'),
+        'password_cache': OptionalValue('--password-cache'),
     })
 
-    def _init(self, config, options):
+    def _init(self):
 
         """Implements _EntryPoint._init."""
 
-        from gi.repository import GObject
         import udiskie.prompt
 
-        mainloop = GObject.MainLoop()
-        daemon = get_backend('Daemon', options['udisks_version'])
+        config = self.config
+        options = self.options
+
+        prompt = udiskie.prompt.password(options['password_prompt'])
         browser = udiskie.prompt.browser(options['file_manager'])
+        cache = None
+
+        if options['password_cache'] is not False:
+            import udiskie.cache
+            timeout = int(options['password_cache']) * 60
+            cache = udiskie.cache.PasswordCache(timeout)
+
         mounter = udiskie.mount.Mounter(
             mount_options=config.mount_options,
             ignore_device=config.ignore_device,
-            prompt=udiskie.prompt.password(True),
+            prompt=prompt,
             browser=browser,
-            udisks=daemon)
+            cache=cache,
+            udisks=self.udisks)
+
+        if options['notify'] and not has_Notify():
+            libnotify_not_available = _(
+                "Typelib for 'libnotify' is not available. Possible causes include:"
+                "\n\t- libnotify is not installed"
+                "\n\t- the typelib is provided by a separate package"
+                "\n\t- libnotify was built with introspection disabled"
+                "\n\nStarting udiskie without notifications.")
+            logging.getLogger(__name__).error(libnotify_not_available)
+            options['notify'] = False
+
+        if options['tray'] and not has_Gtk(3):
+            gtk3_not_available = _(
+                "Typelib for 'Gtk 3.0' is not available. Possible causes include:"
+                "\n\t- GTK3 is not installed"
+                "\n\t- the typelib is provided by a separate package"
+                "\n\t- GTK3 was built with introspection disabled"
+                "\n\nStarting udiskie without tray icon.")
+            logging.getLogger(__name__).error(gtk3_not_available)
+            options['tray'] = False
+
+        tasks = []
 
         # notifications (optional):
         if options['notify']:
             import udiskie.notify
             from gi.repository import Notify
             Notify.init('udiskie')
+            aconfig = config.notification_actions
+            if options['automount']:
+                aconfig.setdefault('device_added', [])
+            else:
+                aconfig.setdefault('device_added', ['mount'])
             udiskie.notify.Notify(Notify.Notification.new,
                                   mounter=mounter,
-                                  timeout=config.notifications)
+                                  timeout=config.notifications,
+                                  aconfig=aconfig)
 
         # tray icon (optional):
         if options['tray']:
@@ -329,33 +398,30 @@ class Daemon(_EntryPoint):
             if options['tray'] not in tray_classes:
                 raise ValueError("Invalid tray: %s" % (options['tray'],))
             icons = udiskie.tray.Icons(config.icon_names)
+            actions = udiskie.mount.DeviceActions(mounter)
             menu_maker = udiskie.tray.SmartUdiskieMenu(
                 mounter,
                 icons,
-                {'quit': mainloop.quit})
+                actions,
+                quit_action=self.mainloop.quit)
             TrayIcon = tray_classes[options['tray']]
             statusicon = TrayIcon(menu_maker, icons)
+            tasks.append(statusicon.task)
         else:
             statusicon = None
+            tasks.append(RunForever)
 
         # automounter
         if options['automount']:
             import udiskie.automount
             udiskie.automount.AutoMounter(mounter)
+            tasks.append(mounter.add_all())
 
         # Note: mounter and statusicon are saved so these are kept alive:
-        self.mainloop = mainloop
         self.mounter = mounter
         self.statusicon = statusicon
 
-    def run(self):
-        """Implements _EntryPoint.run."""
-        if self.options['automount']:
-            self.mounter.add_all()
-        try:
-            return self.mainloop.run()
-        except KeyboardInterrupt:
-            return 0
+        return AsyncList(tasks)
 
 
 class Mount(_EntryPoint):
@@ -388,6 +454,9 @@ class Mount(_EntryPoint):
         -R, --no-recursive                      Disable recursive mounting
 
         -o OPTIONS, --options OPTIONS           Mount option list
+
+        -p COMMAND, --password-prompt COMMAND   Command for password retrieval
+        -P, --no-password-prompt                Disable unlocking
     """
 
     name = 'udiskie-mount'
@@ -396,42 +465,46 @@ class Mount(_EntryPoint):
         'recursive': False,
         'options': None,
         '<device>': None,
+        'password_prompt': 'builtin:tty',
     })
 
     option_rules = extend(_EntryPoint.option_rules, {
         'recursive': Switch('recursive'),
         'options': Value('--options'),
         '<device>': Value('DEVICE'),
+        'password_prompt': OptionalValue('--password-prompt'),
     })
 
-    def _init(self, config, options):
+    def _init(self):
+
         """Implements _EntryPoint._init."""
+
         import udiskie.prompt
+
+        config = self.config
+        options = self.options
+
         if options['options']:
             opts = [o.strip() for o in options['options'].split(',')]
             mount_options = lambda dev: opts
         else:
             mount_options = config.mount_options
-        self.mounter = udiskie.mount.Mounter(
+
+        prompt = udiskie.prompt.password(options['password_prompt'])
+        mounter = udiskie.mount.Mounter(
             mount_options=mount_options,
             ignore_device=config.ignore_device,
-            prompt=udiskie.prompt.password(False),
-            udisks=get_backend('Sniffer', options['udisks_version']))
+            prompt=prompt,
+            udisks=self.udisks)
 
-    def run(self):
-        """Implements _EntryPoint.run."""
-        options = self.options
-        mounter = self.mounter
         recursive = options['recursive']
-        # only mount the desired devices
         if options['<device>']:
-            success = True
-            for path in options['<device>']:
-                success = success and mounter.add(path, recursive=recursive)
-        # mount all present devices
+            tasks = [mounter.add(path, recursive=recursive)
+                     for path in options['<device>']]
+            # TODO: AND results
+            return AsyncList(tasks)
         else:
-            success = mounter.add_all(recursive=recursive)
-        return 0 if success else 1
+            return mounter.add_all(recursive=recursive)
 
 
 class Umount(_EntryPoint):
@@ -491,23 +564,21 @@ class Umount(_EntryPoint):
         '<device>': Value('DEVICE'),
     })
 
-    def _init(self, config, options):
+    def _init(self):
+
         """Implements _EntryPoint._init."""
-        self.mounter = udiskie.mount.Mounter(
-            udisks=get_backend('Sniffer', options['udisks_version']))
 
-    def run(self):
-        """Implements _EntryPoint.run."""
         options = self.options
-        mounter = self.mounter
+        mounter = udiskie.mount.Mounter(self.udisks)
+
         strategy = dict(detach=options['detach'],
                         eject=options['eject'],
                         lock=options['lock'])
         if options['<device>']:
             strategy['force'] = options['force']
-            success = True
-            for path in options['<device>']:
-                success = mounter.remove(path, **strategy) and success
+            tasks = [mounter.remove(path, **strategy)
+                     for path in options['<device>']]
+            # TODO: AND results
+            return AsyncList(tasks)
         else:
-            success = mounter.remove_all(**strategy)
-        return 0 if success else 1
+            return mounter.remove_all(**strategy)
