@@ -5,12 +5,15 @@ For an example config file, see the manual. If you don't have the man page
 installed, a raw version is available in doc/udiskie.8.txt.
 """
 
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
 import logging
 import os
-import sys
 
-from udiskie.compat import basestring
-from udiskie.locale import _
+from .common import exc_message
+from .compat import basestring, fix_str_conversions
+from .locale import _
 
 
 __all__ = ['DeviceFilter',
@@ -25,6 +28,18 @@ def lower(s):
         return s
 
 
+def yaml_load(stream):
+    """Load YAML document, but load all strings as unicode on py2."""
+    import yaml
+    class UnicodeLoader(yaml.SafeLoader):
+        pass
+    UnicodeLoader.add_constructor(
+        yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG,
+        UnicodeLoader.construct_scalar)
+    return yaml.load(stream, UnicodeLoader)
+
+
+@fix_str_conversions
 class DeviceFilter(object):
 
     """Associate a certain value to matching devices."""
@@ -36,6 +51,7 @@ class DeviceFilter(object):
         'is_partition',
         'is_filesystem',
         'is_luks',
+        'is_loop',
         'is_toplevel',
         'is_detachable',
         'is_ejectable',
@@ -56,6 +72,9 @@ class DeviceFilter(object):
         'mount_paths',
         'is_unlocked',
         'in_use',
+        'should_automount',
+        'ui_label',
+        'loop_filename',
     ]
 
     def __init__(self, match, value):
@@ -194,14 +213,12 @@ class Config(object):
             for path in cls.default_pathes():
                 try:
                     return cls.from_file(path)
-                except IOError:
+                except IOError as e:
                     logging.getLogger(__name__).debug(
-                        "Failed to read config file: {0}"
-                        .format(sys.exc_info()[1]))
-                except ImportError:
+                        _("Failed to read config file: {0}", exc_message(e)))
+                except ImportError as e:
                     logging.getLogger(__name__).warn(
-                        "Failed to read {0!r}: {1}"
-                        .format(path, sys.exc_info()[1]))
+                        _("Failed to read {0!r}: {1}", path, exc_message(e)))
             return cls({})
         # False/'' => no config
         if not path:
@@ -209,7 +226,7 @@ class Config(object):
         if os.path.splitext(path)[1].lower() == '.json':
             from json import load
         else:
-            from yaml import safe_load as load
+            load = yaml_load
         with open(path) as f:
             return cls(load(f))
 
@@ -239,3 +256,8 @@ class Config(object):
     def icon_names(self):
         """Get the icon names dictionary from the config file."""
         return self._data.get('icon_names', {}).copy()
+
+    @property
+    def notification_actions(self):
+        """Get the notification actions dictionary from the config file."""
+        return self._data.get('notification_actions', {}).copy()
