@@ -2,11 +2,16 @@
 Notification utility.
 """
 
-import logging
-import sys
+from __future__ import absolute_import
+from __future__ import unicode_literals
 
-from udiskie.dbus import DBusException
-from udiskie.locale import _
+import logging
+
+from gi.repository import GLib
+
+from .common import exc_message
+from .mount import DeviceActions
+from .locale import _
 
 
 __all__ = ['Notify']
@@ -24,7 +29,7 @@ class Notify(object):
     notification services.
     """
 
-    def __init__(self, notify, mounter, timeout=None):
+    def __init__(self, notify, mounter, timeout=None, aconfig=None):
         """
         Initialize notifier and connect to service.
 
@@ -34,7 +39,9 @@ class Notify(object):
         """
         self._notify = notify
         self._mounter = mounter
+        self._actions = DeviceActions(mounter)
         self._timeout = timeout or {}
+        self._aconfig = aconfig or {}
         self._default = self._timeout.get('timeout', -1)
         self._log = logging.getLogger(__name__)
         self._notifications = []
@@ -54,15 +61,13 @@ class Notify(object):
 
         :param device: device object
         """
-        label = device.id_label
-        mount_path = device.mount_paths[0]
         browse_action = ('browse', _('Browse directory'),
                          self._mounter.browse, device)
         self._show_notification(
             'device_mounted',
             _('Device mounted'),
-            _('{0.id_label} mounted on {0.mount_paths[0]}', device),
-            'drive-removable-media',
+            _('{0.ui_label} mounted on {0.mount_paths[0]}', device),
+            device.icon_name,
             self._mounter._browser and browse_action)
 
     def device_unmounted(self, device):
@@ -71,12 +76,11 @@ class Notify(object):
 
         :param device: device object
         """
-        label = device.id_label
         self._show_notification(
             'device_unmounted',
             _('Device unmounted'),
-            _('{0.id_label} unmounted', device),
-            'drive-removable-media')
+            _('{0.ui_label} unmounted', device),
+            device.icon_name)
 
     def device_locked(self, device):
         """
@@ -84,12 +88,11 @@ class Notify(object):
 
         :param device: device object
         """
-        device_file = device.device_presentation
         self._show_notification(
             'device_locked',
             _('Device locked'),
             _('{0.device_presentation} locked', device),
-            'drive-removable-media')
+            device.icon_name)
 
     def device_unlocked(self, device):
         """
@@ -97,12 +100,11 @@ class Notify(object):
 
         :param device: device object
         """
-        device_file = device.device_presentation
         self._show_notification(
             'device_unlocked',
             _('Device unlocked'),
             _('{0.device_presentation} unlocked', device),
-            'drive-removable-media')
+            device.icon_name)
 
     def device_added(self, device):
         """
@@ -110,13 +112,43 @@ class Notify(object):
 
         :param device: device object
         """
+        if self._has_actions('device_added'):
+            # wait for partitions etc to be reported to udiskie, otherwise we
+            # can't discover the actions
+            GLib.timeout_add(500, self._device_added, device)
+        else:
+            self._device_added(device)
+
+    def _device_added(self, device):
         device_file = device.device_presentation
         if (device.is_drive or device.is_toplevel) and device_file:
+            # On UDisks1: cannot invoke self._actions.detect() for newly added
+            # LUKS devices. It should be okay if we had waited for the actions
+            # to be added, though.
+            if self._has_actions('device_added'):
+                node_tree = self._actions.detect(device.object_path)
+                flat_actions = self._flatten_node(node_tree)
+                actions = [
+                    (action.method,
+                     action.label.format(action.device.ui_label),
+                     action.action)
+                    for action in flat_actions
+                ]
+            else:
+                actions = ()
             self._show_notification(
                 'device_added',
                 _('Device added'),
                 _('device appeared on {0.device_presentation}', device),
-                'drive-removable-media')
+                device.icon_name,
+                *actions)
+
+    def _flatten_node(self, node):
+        actions = [action
+                   for branch in node.branches
+                   for action in self._flatten_node(branch)]
+        actions += node.methods
+        return actions
 
     def device_removed(self, device):
         """
@@ -130,7 +162,7 @@ class Notify(object):
                 'device_removed',
                 _('Device removed'),
                 _('device disappeared on {0.device_presentation}', device),
-                'drive-removable-media')
+                device.icon_name)
 
     def job_failed(self, device, action, message):
         """
@@ -152,12 +184,12 @@ class Notify(object):
         self._show_notification(
             'job_failed',
             _('Job failed'), text,
-            'drive-removable-media',
+            device.icon_name,
             retry_action)
 
     def _show_notification(self,
                            event, summary, message, icon,
-                           action=None):
+                           *actions):
         """
         Show a notification.
 
@@ -171,19 +203,18 @@ class Notify(object):
         timeout = self._get_timeout(event)
         if timeout != -1:
             notification.set_timeout(int(timeout * 1000))
-        if action:
-            self._add_action(notification, *action)
+        for action in actions:
+            if self._action_enabled(event, action[0]):
+                self._add_action(notification, *action)
         try:
             notification.show()
-        except DBusException:
+        except GLib.GError as exc:
             # Catch and log the exception. Starting udiskie with notifications
             # enabled while not having a notification service installed is a
             # mistake too easy to be made, but it shoud not render the rest of
             # udiskie's logic useless by raising an exception before the
             # automount handler gets invoked.
-            exc = sys.exc_info()[1]
-            self._log.error("Failed to show notification: {0}"
-                            .format(exc.message))
+            self._log.error(_("Failed to show notification: {0}", exc_message(exc)))
 
     def _add_action(self, notification, action, label, callback, *args):
         """
@@ -228,3 +259,25 @@ class Notify(object):
         :rtype: int, float or NoneType
         """
         return self._timeout.get(event, self._default)
+
+    def _action_enabled(self, event, action):
+        """
+        Check if an action for a notification is enabled.
+
+        :param str event: event name
+        :param str action: action name
+        :rtype: bool
+        """
+        event_actions = self._aconfig.get(event)
+        if event_actions is None:
+            return True
+        if event_actions is False:
+            return False
+        return action in event_actions
+
+    def _has_actions(self, event):
+        """
+        Check if a notification type has any enabled actions.
+        """
+        event_actions = self._aconfig.get(event)
+        return event_actions is None or bool(event_actions)
