@@ -2,27 +2,22 @@
 Tray icon for udiskie.
 """
 
-from collections import namedtuple
-from functools import partial
-from itertools import chain
+from __future__ import absolute_import
+from __future__ import unicode_literals
 
-from gi.repository import Gtk
 from gi.repository import Gio
+from gi.repository import Gtk
 
-from udiskie.common import setdefault
-from udiskie.compat import basestring
-from udiskie.locale import _
+from .async_ import Async
+from .common import setdefault
+from .compat import basestring
+from .locale import _
+from .mount import Action, Branch, prune_empty_node
 
 
 __all__ = ['UdiskieMenu', 'SmartUdiskieMenu', 'TrayIcon']
 
 
-# data structs containing the menu hierarchy:
-Node = namedtuple('Node', ['root', 'branches', 'device', 'label', 'methods'])
-Action = namedtuple('Action', ['label', 'device', 'method'])
-Branch = namedtuple('Branch', ['label', 'groups'])
-
-
 class Icons(object):
 
     """Encapsulates the responsibility to load icons."""
@@ -43,6 +38,7 @@ class Icons(object):
         'eject': ['udiskie-eject', 'media-eject'],
         'detach': ['udiskie-detach'],
         'quit': ['application-exit'],
+        'forget_password': ['edit-delete'],
     }
 
     def __init__(self, icon_names={}):
@@ -50,7 +46,7 @@ class Icons(object):
         _icon_names = icon_names.copy()
         setdefault(_icon_names, self.__class__._icon_names)
         self._icon_names = _icon_names
-        for k,v in _icon_names.items():
+        for k, v in _icon_names.items():
             if isinstance(v, basestring):
                 self._icon_names[k] = [v]
 
@@ -85,27 +81,19 @@ class UdiskieMenu(object):
     Objects of this class generate action menus when being called.
     """
 
-    _menu_labels = {
-        'browse': _('Browse {0}'),
-        'mount': _('Mount {0}'),
-        'unmount': _('Unmount {0}'),
-        'unlock': _('Unlock {0}'),
-        'lock': _('Lock {0}'),
-        'eject': _('Eject {0}'),
-        'detach': _('Unpower {0}'),
-        'quit': _('Quit'), }
+    _quit_label = _('Quit')
 
-    def __init__(self, mounter, icons, actions={}):
+    def __init__(self, mounter, icons, actions, quit_action=None):
         """
         Initialize a new menu maker.
 
         :param object mounter: mount operation provider
         :param Icons icons: icon provider
-        :param dict actions: actions for menu items
+        :param DeviceActions actions: device actions discovery
         :returns: a new menu maker
         :rtype: cls
 
-        Required keys for the ``_menu_labels``, ``_menu_icons`` and
+        Required keys for the ``_labels``, ``_menu_icons`` and
         ``actions`` dictionaries are:
 
             - browse    Open mount location
@@ -122,17 +110,8 @@ class UdiskieMenu(object):
         """
         self._icons = icons
         self._mounter = mounter
-        _actions = actions.copy()
-        setdefault(_actions, {
-            'browse': mounter.browse,
-            'mount': mounter.mount,
-            'unmount': mounter.unmount,
-            'unlock': mounter.unlock,
-            'lock': partial(mounter.remove, force=True),
-            'eject': partial(mounter.eject, force=True),
-            'detach': partial(mounter.detach, force=True),
-            'quit': Gtk.main_quit, })
-        self._actions = _actions
+        self._actions = actions
+        self._quit_action = quit_action
 
     def __call__(self):
         """
@@ -144,10 +123,14 @@ class UdiskieMenu(object):
         # create actions items
         menu = self._branchmenu(self._prepare_menu(self.detect()).groups)
         # append menu item for closing the application
-        if self._actions.get('quit'):
+        if self._quit_action:
             if len(menu) > 0:
                 menu.append(Gtk.SeparatorMenuItem())
-            menu.append(self._actionitem('quit'))
+            menu.append(self._menuitem(
+                self._quit_label,
+                self._icons.get_icon('quit', Gtk.IconSize.MENU),
+                lambda _: self._quit_action()
+            ))
         return menu
 
     def detect(self):
@@ -155,14 +138,10 @@ class UdiskieMenu(object):
         Detect all currently known devices.
 
         :returns: root of device hierarchy
-        :rtype: Node
+        :rtype: Device
         """
-        root = Node(None, [], None, "", [])
-        device_nodes = dict(map(self._device_node,
-                                self._mounter.get_all_handleable()))
-        # insert child devices as branches into their roots:
-        for object_path, node in device_nodes.items():
-            device_nodes.get(node.root, root).branches.append(node)
+        root = self._actions.detect()
+        prune_empty_node(root, set())
         return root
 
     def _branchmenu(self, groups):
@@ -173,6 +152,8 @@ class UdiskieMenu(object):
         :returns: a new menu object holding all groups of the node
         :rtype: Gtk.Menu
         """
+        def make_action_callback(node):
+            return lambda _: node.action()
         menu = Gtk.Menu()
         separate = False
         for group in groups:
@@ -182,10 +163,10 @@ class UdiskieMenu(object):
                 separate = True
             for node in group:
                 if isinstance(node, Action):
-                    menu.append(self._actionitem(
-                        node.method,
-                        feed=[node.label],
-                        bind=[node.device]))
+                    menu.append(self._menuitem(
+                        node.label,
+                        self._icons.get_icon(node.method, Gtk.IconSize.MENU),
+                        make_action_callback(node)))
                 elif isinstance(node, Branch):
                     menu.append(self._menuitem(
                         node.label,
@@ -221,59 +202,11 @@ class UdiskieMenu(object):
             item.connect('activate', onclick)
         return item
 
-    def _actionitem(self, action, feed=(), bind=()):
-        """
-        Create a menu item for the specified action.
-
-        :param str action: name of the action
-        :param tuple feed: parameters for the label text
-        :param tuple bind: parameters for the onclick handler
-        :returns: the menu item object
-        :rtype: Gtk.MenuItem
-        """
-        return self._menuitem(
-            self._menu_labels[action].format(*feed),
-            self._icons.get_icon(action, Gtk.IconSize.MENU),
-            lambda _: self._actions[action](*bind))
-
-    def _device_node(self, device):
-        """Create an empty menu node for the specified device."""
-        label = device.id_label or device.device_presentation
-        # determine available methods
-        methods = []
-        def append(method):
-            if self._actions[method]:
-                methods.append(method)
-        if device.is_filesystem:
-            if device.is_mounted:
-                append('browse')
-                append('unmount')
-            else:
-                append('mount')
-        elif device.is_crypto:
-            if device.is_unlocked:
-                append('lock')
-            else:
-                append('unlock')
-        if device.is_ejectable and device.has_media:
-            append('eject')
-        if device.is_detachable:
-            append('detach')
-        # find the root device:
-        if device.is_partition:
-            root = device.partition_slave.object_path
-        elif device.is_luks_cleartext:
-            root = device.luks_cleartext_slave.object_path
-        else:
-            root = None
-        # in this first step leave branches empty
-        return device.object_path, Node(root, [], device, label, methods)
-
     def _prepare_menu(self, node):
         """
         Prepare the menu hierarchy from the given device tree.
 
-        :param Node node: root node of device hierarchy
+        :param Device node: root node of device hierarchy
         :returns: menu hierarchy
         :rtype: Branch
         """
@@ -281,9 +214,9 @@ class UdiskieMenu(object):
             label=node.label,
             groups=[
                 [self._prepare_menu(branch)
-                 for branch in node.branches],
-                [Action(node.label, node.device, method)
-                 for method in node.methods],
+                 for branch in node.branches
+                 if branch.methods or branch.branches],
+                node.methods,
             ])
 
 
@@ -293,62 +226,69 @@ class SmartUdiskieMenu(UdiskieMenu):
         """
         Create the actions group for the specified device node.
 
-        :param Node node: device
+        :param Device node: device
         :param str presentation: node label
         """
-        return [Action(presentation, node.device, method)
-                for method in node.methods]
+        labels = self._actions._labels
+        return [Action(action.method,
+                       action.device,
+                       labels[action.method].format(presentation),
+                       action.action)
+                for action in node.methods]
 
-    def _leaves_group(self, node, outer_methods, presentation):
-        """
-        Create groups for the specified node.
-
-        :param Node node: device
-        :param list outer_methods: mix-in methods of root device
-        :param str presentation: node label
-        """
-        if not presentation or (node.device.is_mounted or
-                                not node.device.is_luks_cleartext):
+    def _collapse_device(self, node, presentation=""):
+        """Collapse device hierarchy into a flat folder."""
+        if (not presentation
+                or node.device.is_mounted
+                or not node.device.is_luks_cleartext):
             presentation = node.label
-        if node.branches:
-            return chain.from_iterable(
-                self._leaves_group(
-                    branch,
-                    self._actions_group(node, presentation) + outer_methods,
-                    presentation)
-                for branch in node.branches)
-        elif len(node.methods) + len(outer_methods) > 0:
-            return Branch(
-                label=presentation,
-                groups=[list(chain(self._actions_group(node, presentation),
-                                   outer_methods))]),
-        else:
-            return ()
+        groups = [group
+                  for branch in node.branches
+                  for group in self._collapse_device(branch, presentation)
+                  if group]
+        groups.append(self._actions_group(node, presentation))
+        return groups
 
     def _prepare_menu(self, node):
         """Overrides UdiskieMenu._prepare_menu."""
         return Branch(
             label=node.label,
-            groups=[list(self._leaves_group(node, [], ""))])
+            groups=[
+                [Branch(branch.label, self._collapse_device(branch))
+                 for branch in node.branches
+                 if branch.methods or branch.branches],
+            ])
 
 
 class TrayIcon(object):
 
     """Default TrayIcon class."""
 
-    def __init__(self, menumaker, icons, statusicon=None):
+    def __init__(self, menumaker, icons, statusicon=None, show=True):
         """
-        Create and show a simple Gtk.StatusIcon.
+        Create an object managing a tray icon.
+
+        The actual Gtk.StatusIcon is only created as soon as you call show()
+        for the first time. The reason to delay its creation is that the GTK
+        icon will be initially visible, which results in a perceptable
+        flickering.
 
         :param UdiskieMenu menumaker: menu factory
         :param Gtk.StatusIcon statusicon: status icon
         """
         self._icons = icons
-        self._icon = statusicon or self._create_statusicon()
+        self._icon = statusicon
         self._menu = menumaker
         self._conn_left = None
         self._conn_right = None
-        self.show()
+        self.task = Async()
+        menumaker._quit_action = self.destroy
+        if show:
+            self.show()
+
+    def destroy(self):
+        self.show(False)
+        self.task.callback()
 
     def _create_statusicon(self):
         """Return a new Gtk.StatusIcon."""
@@ -364,22 +304,19 @@ class TrayIcon(object):
 
     def show(self, show=True):
         """Show or hide the tray icon."""
-        if show:
-            if not self.visible:
-                self._show()
-        else:
-            self.hide()
-
-    def hide(self):
-        """Hide the tray icon."""
-        if self.visible:
+        if show and not self.visible:
+            self._show()
+        if not show and self.visible:
             self._hide()
 
     def _show(self):
         """Show the tray icon."""
-        self._icon.set_visible(True)
-        self._conn_left = self._icon.connect("activate", self._left_click_event)
-        self._conn_right = self._icon.connect("popup-menu", self._right_click_event)
+        if not self._icon:
+            self._icon = self._create_statusicon()
+        widget = self._icon
+        widget.set_visible(True)
+        self._conn_left = widget.connect("activate", self._activate)
+        self._conn_right = widget.connect("popup-menu", self._popup_menu)
 
     def _hide(self):
         """Hide the tray icon."""
@@ -393,20 +330,11 @@ class TrayIcon(object):
         """Create the context menu."""
         return self._menu()
 
-    def _left_click_event(self, icon):
+    def _activate(self, icon):
         """Handle a left click event (show the menu)."""
-        m = self.create_context_menu()
-        m.show_all()
-        m.popup(parent_menu_shell=None,
-                parent_menu_item=None,
-                func=icon.position_menu,
-                data=icon,
-                button=0,
-                activate_time=Gtk.get_current_event_time())
-        # need to store reference or menu will be destroyed before showing:
-        self._m = m
+        self._popup_menu(icon, button=0, time=Gtk.get_current_event_time())
 
-    def _right_click_event(self, icon, button, time):
+    def _popup_menu(self, icon, button, time):
         """Handle a right click event (show the menu)."""
         m = self.create_context_menu()
         m.show_all()
@@ -435,42 +363,19 @@ class AutoTray(TrayIcon):
 
         Overrides TrayIcon.__init__.
         """
-        # The reason to overwrite TrayIcon.__init__ is that the AutoTray
-        # icon may need to be hidden at initialization time. When creating a
-        # Gtk.StatusIcon, it will initially be visible, creating a minor
-        # nuisance.
-        self._icons = icons
-        self._icon = None
-        self._menu = menumaker
-        self._conn_left = None
-        self._conn_right = None
+        super(AutoTray, self).__init__(menumaker, icons, show=False)
         # Okay, the following is BAD:
-        menumaker._actions['quit'] = None
-        menumaker._mounter.udisks.connect_all(self)
-        self.show(self.has_menu())
-
-    def _show(self):
-        """Extends TrayIcon._show: create a new status icon."""
-        self._icon = self._create_statusicon()
-        super(AutoTray, self)._show()
-
-    def _hide(self):
-        """Extends TrayIcon._hide: forget the status icon."""
-        super(AutoTray, self)._hide()
-        self._icon = None
+        menumaker._quit_action = None
+        udisks = menumaker._mounter.udisks
+        udisks.connect('device_changed', self.update)
+        udisks.connect('device_added', self.update)
+        udisks.connect('device_removed', self.update)
+        self.update()
 
     def has_menu(self):
         """Check if a menu action is available."""
         return any(self._menu._prepare_menu(self._menu.detect()).groups)
 
-    def device_changed(self, old_state, new_state):
-        """Update visibility."""
-        self.show(self.has_menu())
-
-    def device_added(self, device):
-        """Update visibility."""
-        self.show(self.has_menu())
-
-    def device_removed(self, device):
-        """Update visibility."""
+    def update(self, *args):
+        """Show/hide icon depending on whether there are devices."""
         self.show(self.has_menu())
